from .validate_exception import ValidateException
from .list_item_exception import ListItemException
from .nonexistent_id_exception import NonexistentIDException


__all__ = (
    'ValidateException',
    'ListItemException',
    'NonexistentIDException',
)



import hashlib
import binascii
import os


def salted_sha512_hash_password(password):
    salt = hashlib.sha256(os.urandom(60)).hexdigest().encode('ascii')
    pwdhash = hashlib.pbkdf2_hmac(
        'sha512', password.encode('utf-8'), salt, 100000
    )
    pwdhash = binascii.hexlify(pwdhash)
    return (salt + pwdhash).decode('ascii')



from pynsodm.exceptions import ValidateException


class BaseField:
    def __init__(self, **kwargs):
        self._is_index = kwargs.get('is_index', False)
        self._is_field = kwargs.get('is_field', True)
        self._is_primary = kwargs.get('is_primary', False)
        self._is_sensitive = kwargs.get('is_sensitive', False)
        self._is_relation = kwargs.get('is_relation', False)
        self._is_resolver = kwargs.get('is_resolver', False)
        self._is_multiple = kwargs.get('is_multiple', False)
        self._valid = kwargs.get('valid', None)
        self._handler = kwargs.get('handler', None)
        self._value = None
        self._is_modified = False

    @property
    def is_index(self) -> bool: return self._is_index

    @property
    def is_field(self) -> bool: return self._is_field

    @property
    def is_primary(self) -> bool: return self._is_primary

    @property
    def is_modified(self) -> bool: return self._is_modified

    @property
    def is_sensitive(self) -> bool: return self._is_sensitive

    @property
    def is_relation(self) -> bool: return self._is_relation

    @property
    def is_resolver(self) -> bool: return self._is_resolver

    @property
    def is_multiple(self) -> bool: return self._is_multiple

    @property
    def value(self): return self._value

    @value.setter
    def value(self, val):
        if self._valid and not self._valid(val):
            raise ValidateException()

        if self._handler:
            self._value = self._handler(val)
        else:
            self._value = val
        self._is_modified = True

    @property
    def safety_value(self): return self._value

    @safety_value.setter
    def safety_value(self, val):
        if self._valid and not self._valid(val):
            raise ValidateException()

        self._value = val
        self._is_modified = True


import json

from datetime import datetime

from pynsodm.rethinkdb_ext import BaseModel


def encoder(sensitive_fields=False, dt_format='%Y.%m.%d %H:%M:%S.%f'):
    class ModelEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, BaseModel):
                if sensitive_fields:
                    return obj.dictionary
                else:
                    return obj.unsensitive_dictionary
            elif isinstance(obj, datetime):
                return obj.strftime(dt_format)
    return ModelEncoder



from .base_field import BaseField
from .one_to_many_resolver_field import OTMResolver


class OTMRelation(BaseField):
    def __init__(self, relation_class, backfield=None, **kwargs):
        self._relation_class = relation_class
        self._backfield = backfield
        self._resolver = OTMResolver

        kwargs['is_relation'] = True

        BaseField.__init__(self, **kwargs)

    def __set__(self, obj, value):
        self.value = value if isinstance(value, str) else value.id

    def __get__(self, obj, type):
        if not obj:
            return self
        return self._relation_class.get(self.value)

    @property
    def relation_class(self): return self._relation_class

    @property
    def backfield(self): return self._backfield

    @property
    def resolver(self): return self._resolver



import os

from rethinkdb import RethinkDB

from pynsodm.rethinkdb_ext import BaseModel


class Storage:
    def __init__(self, **kwargs):
        self._driver = RethinkDB()
        self._connection = kwargs.get('connection', None)

        self._host = kwargs.get(
            'host', os.environ.get('RETHINKDB_HOST', 'localhost'))
        self._port = int(kwargs.get(
            'port', os.environ.get('RETHINKDB_PORT', '28015')))
        self._user = kwargs.get(
            'user', os.environ.get('RETHINKDB_USER', 'admin'))
        self._password = kwargs.get(
            'password', os.environ.get('RETHINKDB_PASSWORD', ''))
        self._db = kwargs.get(
            'db', os.environ.get('RETHINKDB_DATABASE', 'test'))

        _models = kwargs.get('models', os.environ.get('RETHINKDB_MODELS', ''))
        self._models = [m for m in _models.split(',') if len(m) > 0]

    def _init_db(self):
        if not self._connection:
            self._connection = self._driver.connect(
                host=self._host,
                port=self._port,
                db=self._db,
                user=self._user,
                password=self._password)
        else:
            self._connection.reconnect()

        db_list = self._driver.db_list().run(self._connection)
        if self._db not in db_list:
            self._driver.db_create(self._db).run(self._connection)

    def _init_index(self, table_name, index):
        index_list = self._driver.table(table_name).index_list().run(
            self._connection)
        if index not in index_list:
            self._driver.table(table_name).index_create(index).run(
                self._connection)
        self._driver.table(table_name).index_wait(index).run(
            self._connection)

    def _init_table(self, table_name, indexes=[]):
        table_list = self._driver.table_list().run(self._connection)

        if table_name not in table_list:
            self._driver.table_create(table_name).run(self._connection)

        for index in indexes:
            self._init_index(table_name, index)

    def connect(self):
        self._init_db()

        subclasses = BaseModel.__subclasses__()

        for subclass in subclasses:
            if len(self._models) > 0 and \
                    subclass.get_model_name() not in self._models:
                continue

            table_name = subclass.get_table_name()

            self._init_table(table_name, subclass.get_index_fields())

            subclass.set_storage(self)
            for relation_field in subclass.get_relation_fields():
                relation_field_obj = getattr(subclass, relation_field)
                if relation_field_obj.backfield:
                    setattr(
                        relation_field_obj.relation_class,
                        relation_field_obj.backfield,
                        relation_field_obj.resolver(subclass))

    def reconnect(self):
        self._connection = None
        self.connect()

    def insert(self, data_obj):
        obj_data = data_obj.dictionary
        obj_data.pop('id')

        for rel_field in data_obj.get_relation_fields():
            rel_object = obj_data.get(rel_field)
            obj_data.pop(rel_field)
            obj_data[rel_field] = rel_object.id

        result = self._driver\
            .table(data_obj.get_table_name())\
            .insert(obj_data)\
            .run(self._connection)

        if 'generated_keys' in result and \
                len(result['generated_keys']) == 1:
            return result['generated_keys'][0]

    def update(self, data_obj):
        obj_data = data_obj.modified_dictionary
        obj_data.pop('id')

        for rel_field in data_obj.get_relation_fields():
            rel_object = obj_data.get(rel_field)
            obj_data.pop(rel_field)
            obj_data[rel_field] = rel_object.id

        self._driver\
            .table(data_obj.get_table_name())\
            .filter({'id': data_obj.id})\
            .update(obj_data)\
            .run(self._connection)

    def get(self, table_name, obj_id):
        return self._driver.table(table_name).get(obj_id).run(self._connection)

    def find(self, table_name, fil):
        return self._driver.table(table_name).filter(fil).run(self._connection)

    def delete(self, table_name, fil):
        result = self._driver\
            .table(table_name)\
            .filter(fil)\
            .delete()\
            .run(self._connection)
        if 'deleted' in result and result['deleted'] > 0:
            return True
        return False



import copy

from pynsodm.fields import IDField, DatetimeField
from pynsodm.exceptions import NonexistentIDException


class BaseModel:
    table_name: str = None
    storage = None

    id = IDField()
    created = DatetimeField(is_index=True, is_sensitive=True)
    updated = DatetimeField(is_index=True, is_sensitive=True)

    def __init__(self, **kwargs):
        self._exist_object = False
        self.fields = {}

        for field_name, field_value in self.get_fields_values().items():
            self.fields[field_name] = copy.deepcopy(field_value)

        for resolver_name in self.get_resolver_fields():
            resolver_value = getattr(self.__class__, resolver_name)
            self.fields[resolver_name] = copy.deepcopy(resolver_value)

        for field_name, field_value in kwargs.items():
            if field_name in dir(self):
                setattr(self, field_name, field_value)

    @classmethod
    def from_dictionary(cls, data, sensitive_fields=False):
        obj = cls()
        obj._exist_object = True

        fields_list = cls.get_fields() \
            if sensitive_fields \
            else cls.get_unsensitive_fields()

        for field_name, field_value in data.items():
            if field_name in fields_list:
                setattr(obj, field_name, field_value)

        return obj

    @classmethod
    def get_table_name(cls):
        if cls.table_name:
            return cls.table_name
        return cls.__name__.lower()

    @classmethod
    def get_model_name(cls):
        return cls.__name__

    @classmethod
    def get_fields_values(cls):
        fields = dir(cls)
        result = {}

        for field in fields:
            field_value = getattr(cls, field)

            if 'is_field' in dir(field_value):
                if field_value.is_field:
                    result[field] = field_value
        return result

    @classmethod
    def get_fields(cls):
        return list(cls.get_fields_values().keys())

    @classmethod
    def get_index_fields(cls):
        return [
            k for k, v
            in cls.get_fields_values().items()
            if v.is_index and not v.is_primary]

    @classmethod
    def get_unsensitive_fields(cls):
        return [
            k for k, v
            in cls.get_fields_values().items()
            if not v.is_sensitive]

    @classmethod
    def get_relation_fields(cls):
        return [k for k, v in cls.get_fields_values().items() if v.is_relation]

    @classmethod
    def get_resolver_fields(cls):
        fields = dir(cls)
        result = []

        for field in fields:
            field_value = getattr(cls, field)

            if 'is_resolver' in dir(field_value):
                if field_value.is_resolver:
                    result.append(field)
        return result

    @classmethod
    def get_modified_fields(cls):
        return [k for k, v in cls.get_fields_values().items() if v.is_modified]

    @classmethod
    def get_primary_index(cls):
        primary_indexes = [
            k for k, v
            in cls.get_fields_values().items()
            if v.is_primary]
        if len(primary_indexes) > 0:
            return primary_indexes[0]
        return None

    @classmethod
    def set_storage(cls, value):
        cls.storage = value

    @classmethod
    def get(cls, id):
        data = cls.storage.get(cls.get_table_name(), id)
        if not data:
            raise NonexistentIDException()

        get_obj = cls.from_dictionary(data, sensitive_fields=True)

        for resolver_field in cls.get_resolver_fields():
            resolver_field_obj = getattr(cls, resolver_field)

            parent_class = resolver_field_obj.relation_class
            parent_class_table = parent_class.get_table_name()
            parent_relation_fields = parent_class.get_relation_fields()
            for parent_relation_field in parent_relation_fields:
                parent_relation_field_obj = getattr(
                    parent_class, parent_relation_field)

                if parent_relation_field_obj.relation_class == cls:
                    data = [
                        elem for elem
                        in cls.storage.find(
                            parent_class_table,
                            {parent_relation_field: get_obj.id})]
                    if len(data) > 0:
                        if not resolver_field_obj.is_multiple:
                            parent = parent_class(**dict(data[0]))
                            setattr(get_obj, resolver_field, parent)
                        else:
                            elements = []
                            for row in data:
                                elements.append(parent_class(**dict(row)))
                                setattr(get_obj, resolver_field, elements)

        return get_obj

    @classmethod
    def find(cls, **fil):
        data = cls.storage.find(cls.get_table_name(), fil)

        return [
            cls.from_dictionary(row, sensitive_fields=True) for row
            in data]

    @classmethod
    def delete(cls, **fil):
        return cls.storage.delete(cls.get_table_name(), fil)

    @property
    def dictionary(self):
        return {
            field_name: getattr(self, field_name) for field_name
            in self.get_fields()}

    @property
    def modified_dictionary(self):
        return {
            field_name: getattr(self, field_name) for field_name
            in self.get_modified_fields()}

    @property
    def unsensitive_dictionary(self):
        return {
            field_name: getattr(self, field_name) for field_name
            in self.get_unsensitive_fields()}

    def __str__(self):
        return f'{self.__class__.__name__}: id {self.id}'

    def __setattr__(self, name, value):
        if name in self.__dict__.get('fields', {}):
            self.__dict__['fields'][name].__set__(self, value)
        else:
            self.__dict__[name] = value

    def __getattribute__(self, name):
        if name == '__dict__':
            return object.__getattribute__(self, '__dict__')
        elif name in self.__dict__.get('fields', {}):
            return self.__dict__['fields'][name].__get__(self, None)
        else:
            return object.__getattribute__(self, name)

    def default(self):
        return self.dictionary

    def save(self):
        if not self.id:
            self.id = self.storage.insert(self)
        else:
            self.updated = None
            self.storage.update(self)



import os
import pytz
from datetime import datetime

from .base_field import BaseField

TIMEZONE = os.environ.get('TIMEZONE', 'UTC')


class DatetimeField(BaseField):
    def __init__(self, **kwargs):
        BaseField.__init__(self, **kwargs)

        self._timezone = pytz.timezone(TIMEZONE)

    def __set__(self, obj, value):
        if not value:
            self.value = self._timezone.localize(datetime.utcnow())
        else:
            self.value = self._timezone.localize(value) \
                if not value.tzinfo \
                else value

    def __get__(self, obj, type):
        if not obj:
            return self
        return self.value \
            if self.value \
            else self._timezone.localize(datetime.utcnow())



from .base_field import BaseField
from .id_fields import IDField
from .datetime_field import DatetimeField
from .string_field import StringField
from .list_field import ListField
from .one_to_one_relation_field import OTORelation
from .one_to_one_resolver_field import OTOResolver
from .one_to_many_relation_field import OTMRelation
from .one_to_many_resolver_field import OTMResolver


__all__ = (
    'BaseField',
    'IDField',
    'DatetimeField',
    'StringField',
    'ListField',
    'OTORelation',
    'OTOResolver',
    'OTMRelation',
    'OTMResolver',
)



from .base_field import BaseField


class ListField(BaseField):
    def __init__(self, **kwargs):
        BaseField.__init__(self, **kwargs)

    def __set__(self, obj, value):
        self.value = list(value)

    def __get__(self, obj, type):
        if not obj:
            return self
        return self.value if self.value else []



from .base_field import BaseField


class OTMResolver(BaseField):
    def __init__(self, relation_class, **kwargs):
        self._relation_class = relation_class

        kwargs['is_resolver'] = True
        kwargs['is_multiple'] = True
        kwargs['is_field'] = False

        BaseField.__init__(self, **kwargs)

    def __set__(self, obj, value):
        self.value = value

    def __get__(self, obj, type):
        if not obj:
            return self
        return self.value

    @property
    def relation_class(self):
        return self._relation_class



from pynsodm.exceptions import ListItemException

from .base_field import BaseField


class StringField(BaseField):
    def __init__(self, **kwargs):
        BaseField.__init__(self, **kwargs)
        self._items = kwargs.get('items', [])

    def __set__(self, obj, value):
        str_val = str(value)
        if not obj._exist_object:
            if len(self._items) > 0 and str_val not in self._items:
                raise ListItemException()
            self.value = str_val
        else:
            self.safety_value = str_val

    def __get__(self, obj, type):
        if not obj:
            return self
        return self.value if self.value else ''



from .base_field import BaseField


class IDField(BaseField):
    def __init__(self, **kwargs):
        if 'is_primary' not in kwargs:
            kwargs['is_primary'] = True
            kwargs['is_index'] = True

        if 'is_sensitive' not in kwargs:
            kwargs['is_sensitive'] = True

        BaseField.__init__(self, **kwargs)

    def __set__(self, obj, value):
        self.value = str(value) if value else None

    def __get__(self, obj, type):
        if not obj:
            return self
        return self.value if self.value else None
